// ----------------------------------------------------------------------------
// Copyright 2016-2018 ARM Ltd.
//
// SPDX-License-Identifier: Apache-2.0
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ----------------------------------------------------------------------------
#ifndef MBED_TEST_MODE
#include "mbed.h"
#include "simple-mbed-cloud-client.h"
#include "LittleFileSystem.h"
#include "HTS221Sensor.h"
#include "LPS22HBSensor.h"
#include "LSM6DSLSensor.h"
#include "lis3mdl_class.h"
// #include "VL53L0X.h"

static DevI2C devI2c(PB_11,PB_10);
static HTS221Sensor sen_hum_temp(&devI2c);
static LPS22HBSensor sen_press_temp(&devI2c);
static LSM6DSLSensor sen_acc_gyro(&devI2c,LSM6DSL_ACC_GYRO_I2C_ADDRESS_LOW,PD_11); // low address
static LIS3MDL sen_mag(&devI2c);
// static DigitalOut shutdown_pin(PC_6);
// static VL53L0X sen_distance(&devI2c, &shutdown_pin, PC_7);

#define SENSORS_POLL_INTERVAL 1.0
//#define SEND_ALL_SENSORS

// An event queue is a very useful structure to debounce information between contexts (e.g. ISR and normal threads)
// This is great because things such as network operations are illegal in ISR, so updating a resource in a button's fall() function is not allowed
EventQueue eventQueue;

// Default network interface object. Don't forget to change the WiFi SSID/password in mbed_app.json if you're using WiFi.
NetworkInterface *net;

// Default block device
BlockDevice* bd = BlockDevice::get_default_instance();
SlicingBlockDevice sd(bd, 0, 2*1024*1024);
LittleFileSystem fs("fs", &sd);

// Default User button for GET example
InterruptIn button(BUTTON1);
// Default LED to use for PUT/POST example
DigitalOut led(LED1);

// Declaring pointers for access to Pelion Client resources outside of main()
MbedCloudClientResource *res_button;
MbedCloudClientResource *res_led;

// Additional resources for sensor readings
MbedCloudClientResource *res_humidity;
MbedCloudClientResource *res_temperature;
#ifdef SEND_ALL_SENSORS
MbedCloudClientResource *res_pressure;
MbedCloudClientResource *res_temperature2;
MbedCloudClientResource *res_magnometer;
MbedCloudClientResource *res_accelerometer;
MbedCloudClientResource *res_gyroscope_x;
MbedCloudClientResource *res_gyroscope_y;
MbedCloudClientResource *res_gyroscope_z;
MbedCloudClientResource *res_distance;
#endif /* SEND_ALL_SENSORS */

// When the device is registered, this variable will be used to access various useful information, like device ID etc.
static const ConnectorClientEndpointInfo* endpointInfo;

/**
 * PUT handler
 * @param resource The resource that triggered the callback
 * @param newValue Updated value for the resource
 */
void led_put_callback(MbedCloudClientResource *resource, m2m::String newValue) {
    printf("PUT received, new value: %s\n", newValue.c_str());
    led = atoi(newValue.c_str());
}

/**
 * POST handler
 * @param resource The resource that triggered the callback
 * @param buffer If a body was passed to the POST function, this contains the data.
 *               Note that the buffer is deallocated after leaving this function, so copy it if you need it longer.
 * @param size Size of the body
 */
void led_post_callback(MbedCloudClientResource *resource, const uint8_t *buffer, uint16_t size) {
    printf("POST received. Payload: %s\n", res_led->get_value().c_str());
    led = atoi(res_led->get_value().c_str());
}

/**
 * Button function triggered by the physical button press.
 */
void button_press() {
    int v = res_button->get_value_int() + 1;
    res_button->set_value(v);
    printf("Button clicked %d times\n", v);
}

/**
 * Notification callback handler
 * @param resource The resource that triggered the callback
 * @param status The delivery status of the notification
 */
void button_callback(MbedCloudClientResource *resource, const NoticationDeliveryStatus status) {
    printf("Button notification, status %s (%d)\n", MbedCloudClientResource::delivery_status_to_string(status), status);
}

/**
 * Registration callback handler
 * @param endpoint Information about the registered endpoint such as the name (so you can find it back in portal)
 */
void registered(const ConnectorClientEndpointInfo *endpoint) {
    printf("Registered to Pelion Device Management. Endpoint Name: %s\n", endpoint->internal_endpoint_name.c_str());
    endpointInfo = endpoint;
}

/**
 * Initialize sensors
 */
void sensors_init() {
    uint8_t id;

    // Initialize sensors
    sen_hum_temp.init(NULL);
    sen_press_temp.init(NULL);
    sen_acc_gyro.init(NULL);
    sen_mag.init(NULL);
    // sen_distance.init_sensor(VL53L0X_DEFAULT_ADDRESS);

    /// Call sensors enable routines
    sen_hum_temp.enable();
    sen_press_temp.enable();
    //sen_mag.enable();
    sen_acc_gyro.enable_x();
    sen_acc_gyro.enable_g();

    printf("\033[2J\033[20A");
    printf ("\nSensors configuration:\n");

    sen_hum_temp.read_id(&id);
    printf("HTS221  humidity & temperature    = 0x%X\n", id);
    sen_press_temp.read_id(&id);
    printf("LPS22HB pressure & temperature    = 0x%X\n", id);
    sen_mag.read_id(&id);
    printf("LIS3MDL magnetometer              = 0x%X\n", id);
    sen_acc_gyro.read_id(&id);
    printf("LSM6DSL accelerometer & gyroscope = 0x%X\n", id);

    printf("\n"); ;
}

/**
 * Update sensors and report their values.
 * This function is called periodically.
 */
void sensors_update() {
    float fval1, fval2;
    int32_t axes[3];
    uint32_t uval;

    printf("\n");

    fval1 = fval2 = 0.0;
    sen_hum_temp.get_humidity(&fval1);
    sen_hum_temp.get_temperature(&fval2);
    printf("HTS221 temp:  %7.2f C,  humidity: %7.2f %%\n", fval1, fval2);
    if (endpointInfo) {
        res_humidity->set_value(fval1);
        res_temperature->set_value(fval2);
    }

    fval1 = fval2 = 0.0;
    sen_press_temp.get_pressure(&fval1);
    sen_press_temp.get_temperature(&fval2);
    printf("LPS22HB temp: %7.2f C,  pressure: %7.2f mbar      \n", fval1, fval2);
#ifdef SEND_ALL_SENSORS
    if (endpointInfo) {
        res_pressure->set_value(fval1);
        res_temperature2->set_value(fval2);
    }
#endif /* SEND_ALL_SENSORS */

    sen_mag.get_m_axes(axes);
    printf("LIS3MDL mag:  %7ld x, %7ld y, %7ld z [mgauss]      \n", axes[0], axes[1], axes[2]);

    sen_acc_gyro.get_x_axes(axes);
    printf("LSM6DSL acc:  %7ld x, %7ld y, %7ld z [mg]       \n", axes[0], axes[1], axes[2]);

    sen_acc_gyro.get_g_axes(axes);
    printf("LSM6DSL gyro: %7ld x, %7ld y, %7ld z [mdps]      \n", axes[0], axes[1], axes[2]);
#ifdef SEND_ALL_SENSORS
    if (endpointInfo) {
        res_gyroscope_x->set_value((float)axes[0]);
        res_gyroscope_y->set_value((float)axes[1]);
        res_gyroscope_z->set_value((float)axes[2]);
    }
#endif /* SEND_ALL_SENSORS */

//     if (sen_distance.get_distance(&uval) == VL53L0X_ERROR_NONE) {
//         printf("VL53L0X [mm]:            %6ld\n", uval);
// #ifdef SEND_ALL_SENSORS
//         if (endpointInfo) res_distance->set_value((float)uval);
// #endif /* SEND_ALL_SENSORS */
//     } else {
//         printf("VL53L0X [mm]:                --\n");
// #ifdef SEND_ALL_SENSORS
//         if (endpointInfo) res_distance->set_value(99999.9);
// #endif /* SEND_ALL_SENSORS */
//     }

    printf("\033[6A");
}

int main(void) {
    printf("\nStarting Simple Pelion Device Management Client example\n");

    // If the User button is pressed ons start, then format storage.
    DigitalIn *user_button = new DigitalIn(BUTTON1);
#if TARGET_DISCO_L475VG_IOT01A
    // The user button on DISCO_L475VG_IOT01A works the other way around
    const int PRESSED = 0;
#else
    const int PRESSED = 1;
#endif
    if (user_button->read() == PRESSED) {
        printf("User button is pushed on start. Formatting the storage...\n");
        int storage_status = fs.reformat(&sd);
        if (storage_status != 0) {
            if (sd.erase(0, sd.size()) == 0) {
                if (fs.format(&sd) == 0) {
                    storage_status = 0;
                    printf("The storage reformatted successfully.\n");
                }
            }
        }
        if (storage_status != 0) {
            printf("ERROR: Failed to reformat the storage (%d).\n", storage_status);
        }
    }

    sensors_init();

    // Connect to the internet (DHCP is expected to be on)
    printf("Connecting to the network using Wifi...\n");
    net = NetworkInterface::get_default_instance();

    nsapi_error_t net_status = -1;
    for (int tries = 0; tries < 3; tries++) {
        net_status = net->connect();
        if (net_status == NSAPI_ERROR_OK) {
            break;
        } else {
            printf("Unable to connect to network. Retrying...\n");
        }
    }

    if (net_status != NSAPI_ERROR_OK) {
        printf("ERROR: Connecting to the network failed (%d)!\n", net_status);
        return -1;
    }

    printf("Connected to the network successfully. IP address: %s\n", net->get_ip_address());

    printf("Initializing Pelion Device Management Client...\n");

    // SimpleMbedCloudClient handles registering over LwM2M to Pelion DM
    SimpleMbedCloudClient client(net, bd, &fs);
    int client_status = client.init();
    if (client_status != 0) {
        printf("ERROR: Pelion Client initialization failed (%d)\n", client_status);
        return -1;
    }

    // Creating resources, which can be written or read from the cloud
    res_button = client.create_resource("3200/0/5501", "button_count");
    res_button->set_value(0);
    res_button->methods(M2MMethod::GET);
    res_button->observable(true);
    res_button->attach_notification_callback(button_callback);

    // Sensor resources
    res_temperature = client.create_resource("3303/0/5700", "temperature");
    res_temperature->set_value(0);
    res_temperature->methods(M2MMethod::GET);
    res_temperature->observable(true);

    res_humidity = client.create_resource("3304/0/5700", "humidity");
    res_humidity->set_value(0);
    res_humidity->methods(M2MMethod::GET);
    res_humidity->observable(true);

#ifdef SEND_ALL_SENSORS
    res_pressure = client.create_resource("3323/1/5700", "pressure");
    res_pressure->set_value(0);
    res_pressure->methods(M2MMethod::GET);
    res_pressure->observable(true);

    res_temperature2 = client.create_resource("3303/1/5700", "temperature");
    res_temperature2->set_value(0);
    res_temperature2->methods(M2MMethod::GET);
    res_temperature2->observable(true);

    res_gyroscope_x = client.create_resource("3334/1/5702", "gyroscope_x");
    res_gyroscope_x->set_value(0);
    res_gyroscope_x->methods(M2MMethod::GET);
    res_gyroscope_x->observable(true);

    res_gyroscope_y = client.create_resource("3334/1/5703", "gyroscope_y");
    res_gyroscope_y->set_value(0);
    res_gyroscope_y->methods(M2MMethod::GET);
    res_gyroscope_y->observable(true);

    res_gyroscope_z = client.create_resource("3334/1/5704", "gyroscope_z");
    res_gyroscope_z->set_value(0);
    res_gyroscope_z->methods(M2MMethod::GET);
    res_gyroscope_z->observable(true);

    res_distance = client.create_resource("3330/0/5700", "distance");
    res_distance->set_value(0);
    res_distance->methods(M2MMethod::GET);
    res_distance->observable(true);

    res_led = client.create_resource("3201/0/5853", "led_state");
    res_led->set_value(1);
    res_led->methods(M2MMethod::GET | M2MMethod::PUT);
    res_led->attach_put_callback(led_put_callback);
#endif /* SEND_ALL_SENSORS */

    printf("Initialized Pelion Client. Registering...\n");

    // Callback that fires when registering is complete
    client.on_registered(&registered);

    // Register with Pelion DM
    client.register_and_connect();

    int i = 600; // wait up 60 seconds before attaching sensors and button events
    while (i-- > 0 && !client.is_client_registered()) {
        wait_ms(100);
    }

    button.fall(eventQueue.event(&button_press));

    // The timer fires on an interrupt context, but debounces it to the eventqueue, so it's safe to do network operations
    Ticker timer;
    timer.attach(eventQueue.event(&sensors_update), SENSORS_POLL_INTERVAL);

    // You can easily run the eventQueue in a separate thread if required
    eventQueue.dispatch_forever();
}
#endif
